@@ -2,14 +2,11 @@
 # gnatsweb-site.pl -
 #     Gnatsweb site-specific variables and subroutines.
 #
-# $Id: gnatsweb-site-sente.pl,v 2.2 1999/04/16 22:59:48 kenstir Exp $
+# $Id: gnatsweb-site-sente.pl,v 2.3 1999/05/08 21:51:44 kenstir Exp $
 
 # Name you want in the page banner.  See also &site_start_html.
 $site_banner_text = 'Sente gnatsweb';
-
-# Default domain for email addresses.  This get tacked onto the end of
-# the user name, but can be overridden.
-$site_default_domain = 'senteinc.com';
+$site_banner_text = 'TEST GNATSWEB' if ($ENV{'REMOTE_ADDR'} eq '127.0.0.1');
 
 # site_callback -
 #     Override gnatsweb behavior under certain circumstances.
@@ -41,6 +38,11 @@ sub site_callback {
  Fixed in: 
 ";
     }
+    elsif ($reason eq 'initialize')
+    {
+        # Include changes to the Fix field in the Audit-Trail.
+        $fieldnames{'Fix'} |= $AUDITINCLUDE;
+    }
     elsif ($reason eq 'page_footer')
     {
         # override footer html
