@@ -5,7 +5,7 @@
 # Copyright 1998-1999 - Matt Gerassimoff
 # and Ken Cox <kenstir@senteinc.com>
 #
-# $Id: gnatsweb.pl,v 2.2 1999/04/16 22:59:49 kenstir Exp $
+# $Id: gnatsweb.pl,v 2.3 1999/05/08 21:51:45 kenstir Exp $
 #
 
 #-----------------------------------------------------------------------------
@@ -46,7 +46,6 @@ else
   die("Can't locate 'sendmail'; must set \$site_mailer in gnats-site.pl");
 }
 
-
 # site_callback -
 #
 #     If defined, this subroutine gets called at various times.  The
@@ -55,6 +54,9 @@ else
 #     thing.  Or, it can return a piece of HTML to implement
 #     site-specific behavior or appearance.
 #
+#     Sorry, the reasons are not documented.  Either put a call to
+#     'warn' into your gnats-site.pl file, or search this file for 'cb('.
+#
 # arguments:
 #     $reason - reason for the call.  Each reason is unique.
 #     @args   - additional parameters may be provided in @args.
@@ -85,7 +87,7 @@ use IO::Handle;
 use POSIX;
 
 # get RCS tag as just a number
-($REVISION  = '$Revision: 2.2 $ ') =~ s/.Revision: (.*) ../$1/;
+( $VERSION = '$Revision: 2.3 $ ' ) =~ s/.Revision: (.*) ../$1/;
 
 # width of text fields
 $textwidth = 60;
@@ -94,11 +96,12 @@ $textwidth = 60;
 $gnats_info_top = "http://www.hyperreal.org/info/gnuinfo/index";
 
 # bits in %fieldnames has (set=yes not-set=no)
-$MULTILINE = 1;     # whether field is multi line
-$SENDEXCLUDE = 2;   # whether the send command should exclude the field
-$REASONCHANGE = 4;  # whether change to a field requires reason
-$ENUM = 8;          # whether field should be displayed as enumerated
-$EDITEXCLUDE = 16;  # if set, don't display on edit page
+$MULTILINE    = 1;   # whether field is multi line
+$SENDEXCLUDE  = 2;   # whether the send command should exclude the field
+$REASONCHANGE = 4;   # whether change to a field requires reason
+$ENUM         = 8;   # whether field should be displayed as enumerated
+$EDITEXCLUDE  = 16;  # if set, don't display on edit page
+$AUDITINCLUDE = 32;  # if set, save changes in Audit-Trail
 
 $REPLY_CONT = 1;
 $REPLY_END = 2;
@@ -134,13 +137,16 @@ sub gerror
   my($text) = @_;
   my $prog = $0;
   $prog =~ s@.*/@@;
-  print "<pre>$prog: $text\n</pre>\n";
+  #print "<pre>$prog: $text\n</pre>\n";
+  print "<h2>Error: $text\n</h2>\n";
 }
 
+# Close the client socket and exit.  The exit can be suppressed by:
+#     local($suppress_client_exit) = 1;
 sub client_exit
 {
   close(SOCK);
-  exit();
+  exit() unless defined($suppress_client_exit);
 }
 
 sub server_reply
@@ -276,6 +282,7 @@ sub get_reply
     $text =~ s/\r//g;
     gerror($text);
     client_exit();
+    @rettext = (); # may get here if exit suppressed
   }
   elsif($state == $CODE_GNATS_LOCKED)
   {
@@ -296,6 +303,7 @@ sub get_reply
   {
     gerror("access denied");
     client_exit();
+    @rettext = (); # may get here if exit suppressed
   }
   else
   {
@@ -354,15 +362,21 @@ sub sendpr
   shift(@submitter_id);
 
   print $q->start_form(),
-	"<table>",
-	"<tr><td><b>E-Mail Address:</b><td>",
-	$q->textfield(-name=>'email',
-	              -default=>$email,
-		      -size=>$textwidth),
+	"<table>";
+  print "<tr><td><b>Reporter's email:</b><td>",
+        $q->textfield(-name=>'email',
+                      -default=>$email,
+                      -size=>$textwidth),
 	"<tr><td><b>CC:</b><td>",
 	$q->textfield(-name=>'cc',
 	              -default=>$cc,
-		      -size=>$textwidth);
+		      -size=>$textwidth),
+	"<tr><td><b>Others to notify<br>of updates:</b><td>",
+	$q->textfield(-name=>'X-GNATS-Notify',
+		      -size=>$textwidth),
+        # a blank row, to separate header info from PR info
+        "<tr><td>&nbsp;<td>&nbsp;\n";
+
   foreach (@fieldnames)
   {
     next if ($fieldnames{$_} & $SENDEXCLUDE);
@@ -399,6 +413,7 @@ sub sendpr
                           -size=>$textwidth,
                           -default=>$default);
     }
+    print "\n";
   }
   print "</table>",
 	$q->submit('cmd', 'submit'),
@@ -410,12 +425,57 @@ sub sendpr
   page_end_html($page);
 }
 
+# validate_email_field -
+#     Used by validate_new_pr to check email address fields in a new PR.
+sub validate_email_field
+{
+  my($fieldname, $fieldval, $required) = @_;
+
+  my $blank = ($fieldval =~ /^\s*$/);
+  if ($required && $blank)
+  {
+    return "$fieldname is blank";
+  }
+  if (!$blank && $fieldval !~ /^\s*\S+\@\S+\.\S+\s*$/)
+  {
+    return "'$fieldval' doesn't look like a valid email address (xxx\@xxx.xxx)";
+  }
+  return '';
+}
+
+# validate_new_pr -
+#     Make sure fields have reasonable values before submitting a new PR.
+sub validate_new_pr
+{
+  my(%fields) = @_;
+  my(@errors) = ();
+  my $err;
+
+  # validate email fields
+  $err = validate_email_field('E-mail Address', $fields{'email'}, 'required');
+  push(@errors, $err) if $err;
+  $err = validate_email_field('CC', $fields{'cc'});
+  push(@errors, $err) if $err;
+  $err = validate_email_field('X-GNATS-Notify', $fields{'X-GNATS-Notify'});
+  push(@errors, $err) if $err;
+
+  # validate some other fields
+  push(@errors, "Synopsis is blank")
+        if($fields{'Synopsis'} =~ /^\s*$/);
+  push(@errors, "Release is blank")
+        if($fields{'Release'} =~ /^\s*$/);
+  push(@errors, "Submitter-Id is 'unknown'")
+        if($fields{'Submitter-Id'} eq 'unknown');
+
+  @errors;
+}
+
 sub submitnewpr
 {
   my $page = 'Create PR Results';
   page_start_html($page);
-  my $debug = 0;
 
+  my $debug = 0;
   my(@values, $key);
   my(%fields);
 
@@ -428,41 +488,50 @@ sub submitnewpr
     }
     $fields{$key} = $val;
   }
-  if($fields{'email'} eq "")
+
+  # Make sure the pr is valid.
+  my(@errors) = validate_new_pr(%fields);
+  if (@errors)
   {
     page_heading($page, 'Error');
-    print "<h3>Your problem report has not been sent.</h3>\n";
-    print "You must fill in the E-Mail Address field.\n";
+    print "<h3>Your problem report has not been sent.</h3>\n",
+          "Fix the following problems, then submit the problem report again:",
+          $q->ul($q->li(\@errors));
     return;
   }
-  $fields{'Originator'} = $fields{'Originator'} || $fields{'email'};
-  $text = unparsepr("send", %fields);
 
+  # Supply a default value for Originator
+  $fields{'Originator'} = $fields{'Originator'} || $fields{'email'};
+
+  # Compose the message
+  my $text = unparsepr('send', %fields);
+  $text = <<EOT . $text;
+To: $config{'GNATS_ADDR'}
+CC: $fields{'cc'}
+Subject: $fields{'Synopsis'}
+From: $fields{'email'}
+Reply-To: $fields{'email'}
+X-gnatsweb-version: $VERSION
+X-GNATS-Notify: $fields{'X-GNATS-Notify'}
+
+EOT
+
+  # Allow debugging
   if($debug)
   {
-    print "<h3>debugging</h3>",
+    print "<h3>debugging -- PR NOT SENT</h3>",
           $q->pre($q->escapeHTML($text));
     page_end_html($page);
     return;
   }
 
+  # Send the message
   if(!open(MAIL, "|$site_mailer"))
   {
     page_heading($page, 'Error');
     print "<h3>Error invoking $site_mailer</h3>";
     return;
   }
-  my $to = $GNATS_ADDR;
-  $text = <<EOT . $text;
-To: $to
-CC: $fields{'cc'}
-Subject: $fields{'Synopsis'}
-From: $fields{'email'}
-Reply-To: $fields{'email'}
-X-gnatsweb-version: $REVISION
-
-EOT
-
   print MAIL $text;
   if(!close(MAIL))
   {
@@ -470,6 +539,8 @@ EOT
     print "<h3>Bad pipe to $site_mailer</h3>";
     exit;
   }
+
+  # Give feedback for success
   page_heading($page, 'Problem Report Sent');
   print "Thank you for your report.  It will take a short while for
 your report to be processed.  When it is, you will receive
@@ -509,12 +580,26 @@ sub view
   print " or "                                if (can_edit() && !$viewaudit);
   print $q->submit('cmd', 'view audit-trail') if (!$viewaudit);
 
+  # print a convenient 'mailto' link
+  my $mailto  = interested_parties($pr, %fields);
+  my $subject = "Re: $fields{'Category'}/$pr";
+  print " or ",
+        "<a href=\"mailto:$mailto?Subject=$subject\">",
+        "send email to interested parties</a>";
+
   print $q->hr(),
         "<table>";
+  print "<tr><td><b>Reporter's email:</b><td>",
+        $q->tt($fields{'Reply-To'}),
+	"<tr><td><b>Others to notify<br>of updates:</b><td>",
+	$q->tt($fields{'X-GNATS-Notify'}),
+        # a blank row, to separate header info from PR info
+        "<tr><td>&nbsp;<td>&nbsp;\n";
+
   foreach (@fieldnames)
   {
     next if $_ eq 'Audit-Trail';
-    my $val = CGI::escapeHTML($fields{$_});
+    my $val = $q->escapeHTML($fields{$_});
     my $valign = '';
     if ($fieldnames{$_} & $MULTILINE)
     {
@@ -533,6 +618,11 @@ sub view
   print " or "                                if (can_edit() && !$viewaudit);
   print $q->submit('cmd', 'view audit-trail') if (!$viewaudit);
 
+  # print a convenient 'mailto' link
+  print " or ",
+        "<a href=\"mailto:$mailto?Subject=$subject\">",
+        "send email to interested parties</a>";
+
   print $q->end_form();
 
   # Footer comes before the audit-trail.
@@ -541,7 +631,7 @@ sub view
   if($viewaudit)
   {
     print "<h3>Audit Trail:</h3>\n",
-          $q->pre(CGI::escapeHTML($fields{'Audit-Trail'}));
+          $q->pre($q->escapeHTML($fields{'Audit-Trail'}));
   }
 
   page_end_html($page);
@@ -552,7 +642,7 @@ sub edit
   my $page = 'Edit PR';
   page_start_html($page);
 
-  my $debug = 0;
+  #my $debug = 0;
 
   my($pr) = $q->param('pr');
   if(!$pr)
@@ -564,20 +654,9 @@ sub edit
   page_heading($page, "Edit Problem Report: $pr", 1);
 
   my(%fields) = readpr($pr);
-#  if ($debug)
-#  {
-#    print "<h3>debugging</h3><font size=1><pre>";
-#    foreach my $f (sort keys %fields)
-#    {
-#      printf "%-16s %s\n", $f, CGI::escapeHTML($fields{$f});
-#    }
-#    print "</pre></font><hr>\n";
-#  }
 
-  # 4/12/99 kenstir: "Carl E. Lindberg" <lindberg@BLaCKSMITH.com> says
-  # that his old (gnats 3.2) PR's contain email addrs in the Responsible
-  # field.  This doesn't change anything for my PR's.
-  $fields{'Responsible'} = fix_email_addr($fields{'Responsible'});
+  # Trim Responsible for compatibility.
+  $fields{'Responsible'} = trim_responsible($fields{'Responsible'});
 
   # remove "all" from arrays
   shift(@category);
@@ -601,19 +680,30 @@ sub edit
         "<hr>\n";
 
   print "<table>";
+  print "<tr><td><b>Reporter's email:</b><td>",
+        $q->textfield(-name=>'Reply-To',
+                      -default=>$fields{'Reply-To'},
+                      -size=>$textwidth),
+	"<tr><td><b>Others to notify<br>of updates:</b><td>",
+	$q->textfield(-name=>'X-GNATS-Notify',
+                      -default=>$fields{'X-GNATS-Notify'},
+		      -size=>$textwidth),
+        # a blank row, to separate header info from PR info
+        "<tr><td>&nbsp;<td>&nbsp;\n";
+
   foreach (@fieldnames)
   {
-    next if ($fieldnames{$_} & $EDITEXCLUDE);
+    next if ($fieldnames{$_} && $fieldnames{$_} & $EDITEXCLUDE);
     my $lc_fieldname = field2param($_);
 
-    if ($fieldnames{$_} & $ENUM)
+    if ($fieldnames{$_} && $fieldnames{$_} & $ENUM)
     {
       print "<tr><td><b>$_:</b><td>",
             $q->popup_menu(-name=>$_,
                            -values=>\@$lc_fieldname,
                            -default=>$fields{$_});
     }
-    elsif ($fieldnames{$_} & $MULTILINE)
+    elsif ($fieldnames{$_} && $fieldnames{$_} & $MULTILINE)
     {
       print "<tr><td valign=top><b>$_:</b><td>",
             $q->textarea(-name=>$_,
@@ -628,7 +718,7 @@ sub edit
                           -size=>$textwidth,
                           -default=>$fields{$_});
     }
-    if ($fieldnames{$_} & $REASONCHANGE)
+    if ($fieldnames{$_} && $fieldnames{$_} & $REASONCHANGE)
     {
       print "<tr><td valign=top><b>Reason Changed:</b><td>",
             $q->textarea(-name=>"$_-Why",
@@ -650,7 +740,7 @@ sub edit
   page_footer($page);
 
   print "<h3>Audit-Trail:</h3>\n",
-        $q->pre(CGI::escapeHTML($fields{'Audit-Trail'}));
+        $q->pre($q->escapeHTML($fields{'Audit-Trail'}));
   page_end_html($page);
 }
 
@@ -677,9 +767,12 @@ sub submitedit
   my(%oldfields) = lockpr($pr, $user);
   LOCKED:
   {
+    # Trim Responsible for compatibility.
+    $oldfields{'Responsible'} = trim_responsible($oldfields{'Responsible'});
+
     # Merge %oldfields and CGI params to get %fields.  Not all gnats
-    # fields are present in the CGI params, so the ones which are not
-    # specified default to their old values.
+    # fields have to be present in the CGI params; the ones which are
+    # not specified default to their old values.
     %fields = %oldfields;
     foreach my $key ($q->param)
     {
@@ -692,20 +785,20 @@ sub submitedit
       $fields{$key} = $val;
     }
 
-#    if ($debug)
-#    {
-#      print "<h3>debugging</h3><font size=1><pre>";
-#      foreach my $f (sort keys %fields)
-#      {
-#	printf "%-16s %s\n", $f, CGI::escapeHTML($fields{$f});
-#      }
-#      print "</pre></font><hr>\n";
-#    }
+    if ($debug)
+    {
+      print "<h3>debugging</h3><font size=1><pre>";
+      foreach my $f (sort keys %fields)
+      {
+	printf "%-16s %s\n", $f, $q->escapeHTML($fields{$f});
+      }
+      print "</pre></font><hr>\n";
+    }
 
     # TODO: fix this default; it doesn't seem right
     if($user eq "")
     {
-      $user = $oldfields{'From'};
+      $user = $oldfields{'Reply-To'};
     }
 
     if($fields{'Last-Modified'} ne $oldfields{'Last-Modified'})
@@ -725,22 +818,43 @@ sub submitedit
       print "<h3>Responsible party is '$fields{'Responsible'}', user is '$user'!</h3>\n";
       last LOCKED;
     }
-    
+
+    # If X-GNATS-Notify or Reply-To changed, we need to splice the
+    # change into the envelope.
+    foreach ('Reply-To', 'X-GNATS-Notify')
+    {
+      if($fields{$_} ne $oldfields{$_})
+      {
+        if ($fields{'envelope'} =~ /^$_: /m)
+        {
+          # Replace existing header with new one.
+          $fields{'envelope'} =~ s/^$_: .*$/$_: $fields{$_}/m;
+        }
+        else
+        {
+          # Insert new header at end (blank line).  Keep blank line at end.
+          $fields{'envelope'} =~ s/^$/$_: $fields{$_}\n/m;
+        }
+      }
+    }
+
+    # Leave an Audit-Trail
     foreach (@fieldnames)
     {
       if($_ ne "Audit-Trail")
       {
-        # 3/30/99 kenstir: These two commands screw up the Audit-Trail by
-        # removing some newlines.  This causes the tests to fail.
-        #$fields{$_} =~ s/^\s*$//g;
-        #$oldfields{$_} =~ s/^\s*$//g;
 	if($fields{$_} ne $oldfields{$_})
 	{
-          next unless ($fieldnames{$_} & $REASONCHANGE);
+          next unless ($fieldnames{$_} & $AUDITINCLUDE);
 	  if($fieldnames{$_} & $MULTILINE)
 	  {
-	    $audittrail .= "$_-Changed-From:\n$oldfields{$_}";
-	    $audittrail .= "$_-Changed-To:\n$fields{$_}";
+            # For multitext fields, indent the values.
+            my $tmp = $oldfields{$_};
+            $tmp =~ s/^/    /gm;
+	    $audittrail .= "$_-Changed-From:\n$tmp";
+            $tmp = $fields{$_};
+            $tmp =~ s/^/    /gm;
+	    $audittrail .= "$_-Changed-To:\n$tmp";
 	  }
           else
           {
@@ -763,44 +877,27 @@ sub submitedit
               $audittrail .= "$_-Changed-Why:\n" . $fields{"$_-Why"};
             }
 	  }
-	  #$audittrail .= "\n";
 	}
       }
     }
+    $fields{'Audit-Trail'} = $oldfields{'Audit-Trail'} . $audittrail;
+
     # every good let's continue
     if($ok)
     {
       my $mail_sent = 0;
-      # 4/1/99 kenstir: don't send mail to the editor; they already know
-      #if ($adr = praddr($user))
-      #{
-      #  $mailto{$adr} = 1;
-      #}
-      $mailto{$notify{$fields{'Category'}}} = 1 if($notify{$fields{'Category'}});
-      $adr = $oldfields{'Reply-To'} || $oldfields{'From'};
-      $mailto{$adr} = 1;
-      if($fields{'Responsible'} ne $oldfields{'Responsible'})
-      {
-	if(!defined($adr = praddr($fields{'Responsible'})))
-	{
-          page_heading($page, 'Error');
-	  print "<h3>Cannot get email address for $fields{'Responsible'}</h3>",
-                "New Responsible: $fields{'Responsible'}";
-	  last LOCKED;
-	}
-	$mailto{lc($adr)} = 1;
-      }
-      # 4/12/99 kenstir: Don't warn if you can't get an email address
-      # for the old Responsible party.  It may have been invalid, or
-      # they may have left the company.
+
+      # Get list of people to notify, then add old responsible person.
+      # If that person doesn't exist, don't worry about it.
+      %mailto = interested_parties($pr, %fields);
       if(defined($adr = praddr($oldfields{'Responsible'})))
       {
-        $mailto{lc($adr)} = 1;
+        $mailto{$adr} = 1;
       }
-      $fields{'Audit-Trail'} = $oldfields{'Audit-Trail'} . $audittrail;
-      my($newpr) = unparsepr("", %fields);
+
+      my($newpr) = unparsepr('gnatsd', %fields);
       $newpr =~ s/\r//g;
-      #print $q->pre(CGI::escapeHTML($newpr));
+      #print $q->pre($q->escapeHTML($newpr));
       #last LOCKED; # debug
 
       # Submit the edits.
@@ -812,7 +909,7 @@ sub submitedit
       my($mailto);
       delete $mailto{''};
       $mailto = join(", ", sort(keys(%mailto)));
-      #print pre(CGI::escapeHTML("mailto->$mailto<-\n"));
+      #print pre($q->escapeHTML("mailto->$mailto<-\n"));
       #last LOCKED; # debug
       if($mailto ne "" && $audittrail ne "")
       {
@@ -820,7 +917,7 @@ sub submitedit
 	{
 	  print MAILER "To: $mailto\n";
 	  print MAILER "From: $user\n";
-	  print MAILER "Subject: [gnatsweb] Re: $fields{'Category'}/$fields{'Number'}\n\n";
+	  print MAILER "Subject: Re: $fields{'Category'}/$pr\n\n";
           if ($oldfields{'Synopsis'} eq $fields{'Synopsis'})
           {
             print MAILER "Synopsis: $fields{'Synopsis'}\n\n";
@@ -832,7 +929,7 @@ sub submitedit
           }
           print MAILER "$audittrail\n";
           # Print URL so that HTML-enabled mail readers can jump to the PR.
-          print MAILER $q->url() . "?cmd=view&pr=$fields{'Number'}\n";
+          print MAILER $q->url() . "?cmd=view&database=$database&pr=$fields{'Number'}\n";
           if(!close(MAILER))
 	  {
             page_heading($page, 'Error');
@@ -1272,7 +1369,7 @@ sub display_query_results
     print "<td nowrap>", nonempty($release)      if $fields{'release'};
     print "<td nowrap>", nonempty($target)       if($site_release_based
                                                     && $fields{'target'});
-    print "<td nowrap>", nonempty($keywords)     if($site_release_based
+    print "<td>", nonempty($keywords)            if($site_release_based
                                                     && $fields{'keywords'});
     print "<td nowrap>$resp"                     if $fields{'responsible'};
     print "<td nowrap>", nonempty($sub)          if $fields{'submitter_id'};
@@ -1363,7 +1460,7 @@ sub print_stored_queries
   {
     if ($cookie =~ /gnatsweb-query-(.*)/)
     {
-      $stored_queries{$1} = $q->cookie($1);
+      $stored_queries{$1} = $q->cookie($cookie);
     }
   }
   if (%stored_queries)
@@ -1494,7 +1591,7 @@ sub main_page
   print '</table>';
   page_footer($page);
   print '<hr><small>',
-        "Gnatsweb $REVISION, brought to you by<br>",
+        "Gnatsweb $VERSION, brought to you by<br>",
         $q->escapeHTML(
                        'Matt Gerassimoff <mg@digalogsys.com> ' .
                        'and Kenneth H. Cox <kenstir@senteinc.com>.'),
@@ -1508,6 +1605,9 @@ sub main_page
 #
 # usage:
 #     $something = cb($reason, @args) || 'default_value';
+#     # -or-
+#     $something = cb($reason, @args)
+#     $something = 'default_value' unless defined($something);
 #
 # arguments:
 #     $reason - reason for the call.  Each reason is unique.
@@ -1700,11 +1800,93 @@ sub fix_multiline_val
   $val;
 }
 
+# parse_config -
+#     Parse the config file, storing the name/value pairs in the global
+#     hash %config.
+sub parse_config
+{
+  my(@lines) = @_;
+
+  %config = ();
+
+  # Default value for GNATS_ADDR is 'bugs'.
+  $config{'GNATS_ADDR'} = 'bugs';
+
+  # Note that the values may be quoted, as the config file uses
+  # Bourne-shell syntax.
+  foreach $_ (@lines)
+  {
+    if (/(\S+)\s*=\s*['"]?([^'"]*)['"]?/)
+    {
+      $config{$1} = $2;
+    }
+  }
+}
+
+# parse_categories -
+#     Parse the categories file.
+sub parse_categories
+{
+  my(@lines) = @_;
+
+  @category = ("all");
+  %category_notify = ();
+
+  foreach $_ (sort @lines)
+  {
+    my($cat, $desc, $resp, $notify) = split(/:/);
+    # Exclude administrative category 'pending'.
+    next if($cat eq 'pending');
+    push(@category, $cat);
+    $category_notify{$cat} = $notify;
+  }
+}
+
+# parse_submitters -
+#     Parse the submitters file.
+sub parse_submitters
+{
+  my(@lines) = @_;
+
+  @submitter_id = ("all");
+  %submitter_contact = ();
+  %submitter_notify = ();
+
+  foreach $_ (sort @lines)
+  {
+    my($submitter, $full_name, $type, $response_time, $contact, $notify)
+          = split(/:/);
+    push(@submitter_id, $submitter);
+    $submitter_contact{$submitter} = $contact;
+    $submitter_notify{$submitter} = $notify;
+  }
+}
+
+# parse_responsible -
+#     Parse the responsible file.
+sub parse_responsible
+{
+  my(@lines) = @_;
+
+  @responsible = ("all");
+  %responsible_fullname = ();
+  %responsible_address = ();
+
+  foreach $_ (sort @lines)
+  {
+    my($person, $fullname, $address) = split(/:/);
+    push(@responsible, $person);
+    $responsible_fullname{$person} = $fullname;
+    $responsible_address{$person} = $address || $person;
+  }
+}
+
 sub initialize
 {
+  my $regression_testing = shift;
+
   @severity = ("all", "critical", "serious", "non-critical");
   @priority = ("all", "high", "medium", "low");
-  @category = ("all");
   @confidential = ("all", "no", "yes");
 
   # @fields - param names of columns displayable in query results
@@ -1745,6 +1927,7 @@ sub initialize
     "Audit-Trail",
     "Unformatted",
   );
+
   # %fieldnames maps the field name to a flag value composed of bits.
   # See $MULTILINE above for bit definitions.
   %fieldnames = (
@@ -1754,8 +1937,8 @@ sub initialize
     "Confidential"  => $ENUM,
     "Severity"      => $ENUM,
     "Priority"      => $ENUM,
-    "Responsible"   => $ENUM | $REASONCHANGE | $SENDEXCLUDE,
-    "State"         => $ENUM | $REASONCHANGE | $SENDEXCLUDE,
+    "Responsible"   => $ENUM | $REASONCHANGE | $SENDEXCLUDE | $AUDITINCLUDE,
+    "State"         => $ENUM | $REASONCHANGE | $SENDEXCLUDE | $AUDITINCLUDE,
     "Target"        => 0,
     "Keywords"      => 0,
     "Date-Required" => 0,
@@ -1775,6 +1958,7 @@ sub initialize
     "Audit-Trail"   => $MULTILINE | $SENDEXCLUDE | $EDITEXCLUDE,
     "Unformatted"   => $MULTILINE | $SENDEXCLUDE | $EDITEXCLUDE,
   );
+
   # gnatsd query commands: maps param name to gnatsd command
   %gnatsd_query = (
     "category"        => 'catg',
@@ -1809,6 +1993,7 @@ sub initialize
     @fieldnames = grep(!/Target|Keywords|Date-Required/, @fieldnames);
   }
 
+  my(@lines);
   my $response;
 
   # Get gnatsd version from initial server connection text.
@@ -1819,84 +2004,104 @@ sub initialize
     $GNATS_VERS = $1;
   }
 
-  client_cmd("chdb $database");
-
-  # Get user permission level from user command.
-  ($response) = client_cmd("user $user $password");
-  $access_level = 'edit';
-  if ($response =~ /User access level set to (\w*)/)
+  # Suppress fatal exit while issuing CHDB and USER commands.  Otherwise
+  # an error in the $user or $database cookie values can cause a user to
+  # get in a bad state.
+  LOGIN:
   {
-    $access_level = $1;
-  }
+    local($suppress_client_exit) = 1
+          unless $regression_testing;
 
-  # Get GNATS_ADDR by reading config file.  Note that its value may be
-  # quoted, as the config file uses Bourne-shell syntax.  The default
-  # value is 'bugs'.
-  $GNATS_ADDR = 'bugs';
-  foreach $_ (client_cmd("lcfg"))
-  {
-    if (/GNATS_ADDR\s*=\s*['"]?([^'"]*)['"]?/)
+    # Issue CHDB command; revert to login page if it fails.
+    ($response) = client_cmd("chdb $database");
+    if (!$response)
     {
-      $GNATS_ADDR = $1;
+      login();
+      exit();
+    }
+
+    # Get user permission level from USER command.  Revert to the
+    # login page if the command fails.
+    ($response) = client_cmd("user $user $password");
+    if (!$response)
+    {
+      login();
+      exit();
+    }
+    $access_level = 'edit';
+    if ($response =~ /User access level set to (\w*)/)
+    {
+      $access_level = $1;
     }
   }
 
-  my @tmp;
-  @tmp = client_cmd("lcat");
-  foreach $_ (sort @tmp)
-  {
-    my($cat, $desc, $resp, $notify) = split(/:/);
-    push(@category, $cat);
-    $notify{$cat} = $notify;
-  }
+  # Get some enumerated lists
+  my($x, $dummy);
   @state = ("all");
   foreach $_ (client_cmd("lsta"))
   {
     ($x, $dummy) = split(/:/);
     push(@state, $x);
   }
-  @responsible = ("all");
-  @tmp = client_cmd("lres");
-  foreach $_ (sort @tmp)
-  {
-    ($x, $dummy) = split(/:/);
-    push(@responsible, $x);
-  }
   @class = ("all");
   foreach $_ (client_cmd("lcla"))
   {
     ($x, $dummy) = split(/:/);
     push(@class, $x);
   }
-  @submitter_id = ("all");
-  @tmp = client_cmd("lsub");
-  foreach $_ (sort @tmp)
-  {
-    ($x, $dummy) = split(/:/);
-    push(@submitter_id, $x);
-  }
+
+  # List various gnats-adm files, and parse their contents for data we
+  # will need later.  Each parse subroutine stashes information away in
+  # its own global vars.  The call to client_cmd() happens here to
+  # enable regression testing of the parse subs using fixed files.
+  @lines = client_cmd("lcfg");
+  parse_config(@lines);
+  @lines = client_cmd("lcat");
+  parse_categories(@lines);
+  @lines = client_cmd("lsub");
+  parse_submitters(@lines);
+  @lines = client_cmd("lres");
+  parse_responsible(@lines);
 
   # Now that everything's all set up, let the site_callback have at it.
   # It's return value doesn't matter, but here it can muck with our defaults.
   cb('initialize');
 }
 
-# fix_email_addr -
-#     Trim email addresses as it appears in an email From or Reply-To
-#     header (and sometimes in the Responsible field of old PR's), into
-#     something we can use to call sendmail with.
+# trim_responsible -
+#     Trim the value of the Responsible field to get a
+#     valid responsible person.  This exists here, and in gnats itself
+#     (modify_pr(), check_pr(), gnats(), append_report()), for
+#     compatibility with old databases, which had 'person (Full Name)'
+#     in the Responsible field.
+sub trim_responsible {
+  my $resp = shift;
+  $resp =~ s/ .*//;
+  $resp;
+}
+
+# fix_email_addrs -
+#     Trim email addresses as they appear in an email From or Reply-To
+#     header into a comma separated list of just the addresses.
 #
-#     Delete everything inside parenthesis and outside <>'s, inclusive.
+#     Delete everything inside ()'s and outside <>'s, inclusive.
 #
-sub fix_email_addr
+sub fix_email_addrs
 {
-  my $arg = shift;
-  $arg = lc($arg);
-  $arg =~ s/\(.*\)//;
-  $arg =~ s/.*<(.*)>.*/$1/;
-  $arg =~ s/^\s+//;
-  $arg =~ s/\s+$//;
-  $arg;
+  my $addrs = shift;
+  my @addrs = split(/,/, $addrs);
+  my @trimmed_addrs;
+  my $addr;
+  foreach $addr (@addrs)
+  {
+    $addr =~ s/\(.*\)//;
+    $addr =~ s/.*<(.*)>.*/$1/;
+    $addr =~ s/^\s+//;
+    $addr =~ s/\s+$//;
+    push(@trimmed_addrs, $addr);
+  }
+  $addrs = join(', ', @trimmed_addrs);
+  $addrs;
 }
 
 sub parsepr
@@ -1939,41 +2144,50 @@ sub parsepr
     }
 
     # Grab a few fields out of the envelope as it flies by
-    if($hdr eq "Reply-To" || $hdr eq "From")
+    if($hdr eq "Reply-To" || $hdr eq "From" || $hdr eq "X-GNATS-Notify")
     {
-      $arg = fix_email_addr($arg);
-      print "<h3>error: internal whitespace in Reply-to: or From: header!</h3>"
-            if ($arg =~ /\s/);
+      $arg = fix_email_addrs($arg);
       $fields{$hdr} = $arg;
       #print "storing, hdr = $hdr, arg = $arg\n";
     }
   }
+
+  # 5/8/99 kenstir: To get the reporter's email address, only
+  # $fields{'Reply-to'} is consulted.  Initialized it from the 'From'
+  # header if it's not set, then discard the 'From' header.
+  $fields{'Reply-To'} = $fields{'Reply-To'} || $fields{'From'};
+  delete $fields{'From'};
+
   # 3/30/99 kenstir: For some reason Unformatted always ends up with an
   # extra newline here.
   $fields{'Unformatted'} =~ s/\n$//m;
+
   return %fields;
 }
 
 # unparsepr -
-#     Turn PR hash back into a string.
-#     'send' arg is used when creating a new PR only.
+#     Turn PR fields hash into a multi-line string.
+#
+#     The $purpose arg controls how things are done.  The possible values
+#     are:
+#         'send'    - PR will be submitted as a new PR via email
+#         'gntasd'  - PR will be filed using gnatsd; proper '.' escaping done
+#         'test'    - we're being called from the regression tests
 sub unparsepr
 {
-  my($send, %fields) = @_;
+  my($purpose, %fields) = @_;
   my($tmp, $text);
   $text = $fields{'envelope'};
   foreach (@fieldnames)
   {
-    next if($send eq "send" && $fieldnames{$_} & $SENDEXCLUDE);
+    next if($purpose eq "send" && $fieldnames{$_} & $SENDEXCLUDE);
     if($fieldnames{$_} & $MULTILINE)
     {
       # Lines which begin with a '.' need to be escaped by another '.'
-      # unless we're using sendmail to send a new PR.
-      # TODO: When gnatsd supports creating a new PR, then we'll need to
-      # do the escape again.
+      # if we're feeding it to gnatsd.
       $tmp = $fields{$_};
       $tmp =~ s/^[.]/../gm
-            unless ($send eq 'send');
+            if ($purpose eq 'gnatsd');
       $text .= sprintf(">$_:\n%s", $tmp);
     }
     else
@@ -2006,19 +2220,62 @@ sub readpr
   return parsepr(client_cmd("full $pr"));
 }
 
+# interested_parties -
+#     Get list of parties to notify about a PR change.
+#
+#     Returns hash in array context; string of email addrs otherwise.
+sub interested_parties
+{
+  my($pr, %fields) = @_;
+
+  # Gnats 3.110 has some problems in MLPR --
+  # * it includes the category's responsible person (even if that person
+  #   is not responsible for this PR)
+  # * it does not include the PR's responsible person
+  # * it does not include the Reply-To or From
+  #
+  # So for now, don't use it.  However, for versions after 3.110 my
+  # patch to the MLPR command should be there and this can be fixed.
+
+  my(@people);
+  my $person;
+
+  ## Get list from MLPR command.
+  #@people = client_cmd("mlpr $pr");
+  # Ignore intro message
+  #@people = grep(!/Addresses to notify/, @people);
+
+  # Get list of people by constructing it ourselves.
+  @people = ();
+  foreach $person ($fields{'Reply-To'},
+                   $fields{'Responsible'},
+                   $fields{'X-GNATS-Notify'},
+                   $category_notify{$fields{'Category'}},
+                   $submitter_contact{$fields{'Submitter-Id'}},
+                   $submitter_notify{$fields{'Submitter-Id'}})
+  {
+    push(@people, $person) if $person;
+  }
+
+  # Expand any unexpanded addresses, and build up the %addrs hash.
+  my(%addrs) = ();
+  my $addr;
+  foreach $person (@people)
+  {
+    $addr = praddr($person) || $person;
+    $addrs{$addr} = 1;
+  }
+  return wantarray ? %addrs : join(', ', keys(%addrs));
+}
+
+# praddr -
+#     Return email address of responsible person, or undef if not found.
 sub praddr
 {
-  my($snick) = shift;
-  my($nick, $long, $adr,  );
-  foreach $_ (client_cmd("lres"))
-  {
-    ($nick, $long, $adr) = split(/:/);
-    if($nick eq $snick)
-    {
-      return $adr;
-    }
-  }
-  return undef;
+  my $person = shift;
+  # Done this way to avoid -w warning
+  my $addr = exists($responsible_address{$person})
+        ? $responsible_address{$person} : undef;
 }
 
 sub login
@@ -2058,8 +2315,10 @@ sub login
 # callable from gnatsweb.
 sub main
 {
-  # Load gnatsweb-site.pl if present.
+  # Load gnatsweb-site.pl if present.  Die if there are errors;
+  # otherwise the person who wrote gnatsweb-site.pl will never know it.
   do 'gnatsweb-site.pl' if (-e 'gnatsweb-site.pl');
+  die $@ if $@;
 
   $q = new CGI;
   $sn = $q->script_name;
@@ -2082,6 +2341,18 @@ sub main
     exit();
   }
 
+  # Print all our cookies
+  #$i = 0;
+  #foreach my $y ($q->cookie())
+  #{
+  #  @c = $q->cookie($y);
+  #  warn "debug 0: $y: @c:\n";
+  #  $i += length($y);
+  #}
+  #@c = $q->raw_cookie();
+  #warn "debug 0.5: @c:\n";
+  #warn "debug 0.5: total size of raw cookies: ", length("@c"), "\n";
+
   # Retrieve the gnatsweb cookie sent by the browser.
   %cookie = $q->cookie('gnatsweb');
   $user     = $cookie{'user'};
@@ -2091,8 +2362,9 @@ sub main
   $cc       = $cookie{'cc'};
   $submitter_id = $cookie{'Submitter-Id'};
   $originator   = $cookie{'Originator'};
-  @columns  = split(' ', $cookie{'columns'});
-  #foreach $key (sort keys %cookie) { warn "debug 1: $key = $cookie{$key}\n";}
+  @columns      = defined($cookie{'columns'})
+        ? split(' ', $cookie{'columns'}) : ();
+  #foreach $key (sort keys %cookie) {warn "debug 1: $key = $cookie{$key}\n";}
 
   # Upon login, store user/password/database in the gnatsweb cookie.
   if($cmd eq 'login')
@@ -2114,7 +2386,7 @@ sub main
     #$cookie{'Category'}    = $Category    = $q->param('Category');
     #$cookie{'Release'}     = $Release     = $q->param('Release');
     #$cookie{'Environment'} = $Environment = $q->param('Environment');
-    #foreach $key (sort keys %cookie) { warn "debug 2: $key = $cookie{$key}\n";}
+    #foreach $key (sort keys %cookie) {warn "debug 2: $key = $cookie{$key}\n";}
   }
 
   # Upon a 'submit query', store column display list in the cookie.
@@ -2129,7 +2401,7 @@ sub main
   $new_cookie = $q->cookie(-name => 'gnatsweb', 
                            -value => \%cookie,
                            -expires => '+1y');
-  #warn "debug 3:\n$new_cookie\n";
+  #warn "debug 3: new_cookie = $new_cookie\n";
 
   # Serve the cookie.
   print $q->header(-cookie=>$new_cookie);
